
/*
 * error.c
 *   The functions in this file represent the libpqtypes error
 *   system.  It offers the API user the ability to set/get errors.
 *   The error system uses a per-thread global error, implemented
 *   using TLS (via declspec(thread) or pthread TLS).  For
 *   non-PQT_THREAD_SAFE builds, a static buffer is used.
 *
 * Copyright (c) 2009 eSilo, LLC. All rights reserved.
 * This is free software; see the source for copying conditions.  There is
 * NO warranty; not even for MERCHANTABILITY or  FITNESS FOR A  PARTICULAR
 * PURPOSE.
 */

#include "libpqtypes-int.h"

#define NUM_ERROR_FIELDS 12
typedef struct
{
	int size;
	char *msg;
	char *errFields[NUM_ERROR_FIELDS];
	char *errFieldBuffer;
	size_t errFieldBufferLen;
} pqterr_t;

/* non thread-safe, use a static */
#ifndef PQT_THREAD_SAFE
static pqterr_t lasterr = {0};

/* MSVC always uses __declspec(thread) */
#elif defined(PQT_MSVC)
static __declspec(thread) pqterr_t lasterr = {0};

/* Unix-variant, MinGW or Cygwin. */
#else
#include <pthread.h>
static pthread_key_t tlskey_lasterr;

/* When the thread dies, this will get called. */
static void tls_free_lasterr(void *value)
{
	if (value)
	{
		pqterr_t *e = (pqterr_t *) value;

		if (e->msg)
			free(e->msg);

		if (e->errFieldBuffer)
			free(e->errFieldBuffer);

		free(value);
		pthread_setspecific(tlskey_lasterr, NULL);

	}
}

/* called before main.  This attribute is available since gcc 2.7.0.
 * It replaced the obsolete _init ... destructor replaced _fini.
 * We used to use pthread_once, but solaris doesn't support this
 * in older versions (its simply a stub returning 0, ouch!).
 */
void __attribute__ ((constructor)) my_init(void)
{
	pthread_key_create(&tlskey_lasterr, tls_free_lasterr);
}
#endif

static void
vseterror(const char *format, va_list ap, int append);

static pqterr_t *
geterr(void)
{
	pqterr_t *err = NULL;

	/* Non thread-safe mode or windows msvc. */
#if !defined(PQT_THREAD_SAFE) || defined(PQT_MSVC)
	err = &lasterr;

	/* systems requiring pthread TLS keys */
#else

	/* System is using FSU Threads, like SCO OpenServer 5, which
	 * has minor prototype differences.
	 */
#	ifdef PTHREAD_FSU
		pthread_getspecific(tlskey_lasterr, (void **) &err);
#	else
		err = (pqterr_t *) pthread_getspecific(tlskey_lasterr);
#	endif

	if (!err)
	{
		err = (pqterr_t *) malloc(sizeof(pqterr_t));
		if (!err)
			return NULL;
		memset(err, 0, sizeof(pqterr_t));
		pthread_setspecific(tlskey_lasterr, err);
	}
#endif

	if (err->size == 0)
	{
		err->size = 256;
		err->msg = (char *) malloc(err->size);
		*err->msg = 0;
	}

	return err;
}

char *
PQgeterror(void)
{
	static char _empty[1] = {0};
	pqterr_t *err = geterr();
	return err ? err->msg : _empty;
}

void
PQseterror(const char *format, ...)
{
	/* clear error message by passing in NULL, PQseterror(NULL).*/
	if (!format)
	{
		pqterr_t *err = geterr();
		if (err)
		{
			*err->msg = 0;
			memset(err->errFields, 0, sizeof(err->errFields));
		}
	}
	else
	{
		va_list ap;
		va_start(ap, format);
		vseterror(format, ap, FALSE);
	}
}

char *
PQgetErrorField(int fieldcode)
{
	pqterr_t *err = geterr();

	if (!err)
		return NULL;

	switch (fieldcode)
	{
		case PG_DIAG_SEVERITY:
			return err->errFields[0];

		case PG_DIAG_SQLSTATE:
			return err->errFields[1];

		case PG_DIAG_MESSAGE_PRIMARY:
			return err->errFields[2];

		case PG_DIAG_MESSAGE_DETAIL:
			return err->errFields[3];

		case PG_DIAG_MESSAGE_HINT:
			return err->errFields[4];

		case PG_DIAG_STATEMENT_POSITION:
			return err->errFields[5];

		case PG_DIAG_INTERNAL_POSITION:
			return err->errFields[6];

		case PG_DIAG_INTERNAL_QUERY:
			return err->errFields[7];

		case PG_DIAG_CONTEXT:
			return err->errFields[8];

		case PG_DIAG_SOURCE_FILE:
			return err->errFields[9];

		case PG_DIAG_SOURCE_LINE:
			return err->errFields[10];

		case PG_DIAG_SOURCE_FUNCTION:
			return err->errFields[11];

		default:
			return NULL;
	}
}

/* Used by pqt_setresultfields */
#define GET_ERROR_FIELD(idx, name) do{ \
	if ((errFields[idx].val  = PQresultErrorField(res, name))) \
	{ \
		errFields[idx].len = strlen(errFields[idx].val) + 1; \
		errFieldBufferLen += errFields[idx].len; \
	} \
}while (0)

void
pqt_setresultfields(const PGresult *res)
{
	int i;
	char *errFieldBuffer;
	size_t errFieldBufferLen = 0;
	pqterr_t *err = geterr();
	struct {
		char *val;
		size_t len;
	} errFields[NUM_ERROR_FIELDS];

	if (!err)
		return;

	/* clear error fields */
	memset(err->errFields, 0, sizeof(err->errFields));

	/* nothing to copy */
	if (!res)
		return;

	GET_ERROR_FIELD(0,  PG_DIAG_SEVERITY);
	GET_ERROR_FIELD(1,  PG_DIAG_SQLSTATE);
	GET_ERROR_FIELD(2,  PG_DIAG_MESSAGE_PRIMARY);
	GET_ERROR_FIELD(3,  PG_DIAG_MESSAGE_DETAIL);
	GET_ERROR_FIELD(4,  PG_DIAG_MESSAGE_HINT);
	GET_ERROR_FIELD(5,  PG_DIAG_STATEMENT_POSITION);
	GET_ERROR_FIELD(6,  PG_DIAG_INTERNAL_POSITION);
	GET_ERROR_FIELD(7,  PG_DIAG_INTERNAL_QUERY);
	GET_ERROR_FIELD(8,  PG_DIAG_CONTEXT);
	GET_ERROR_FIELD(9,  PG_DIAG_SOURCE_FILE);
	GET_ERROR_FIELD(10, PG_DIAG_SOURCE_LINE);
	GET_ERROR_FIELD(11, PG_DIAG_SOURCE_FUNCTION);

	/* grow memory errFieldBuffer */
	if (errFieldBufferLen > err->errFieldBufferLen)
	{
		errFieldBuffer = (char *) pqt_realloc(
			err->errFieldBuffer, errFieldBufferLen);

		if (!errFieldBuffer)
			return;

		err->errFieldBuffer = errFieldBuffer;
		err->errFieldBufferLen = errFieldBufferLen;
	}
	else
	{
		errFieldBuffer = err->errFieldBuffer;
	}

	for (i=0; i < NUM_ERROR_FIELDS; i++)
	{
		if (errFields[i].val)
		{
			memcpy(errFieldBuffer, errFields[i].val, errFields[i].len);
			err->errFields[i] = errFieldBuffer;
			errFieldBuffer += errFields[i].len;
		}
	}
}

/* errorf() callback for PGtypeArgs, see PQputf() and PQgetf().
 * Always returns -1.
 */
int
pqt_argserrorf(PGtypeArgs *args, const char *format, ...)
{
	va_list ap;
	char fqtn[200];

	if (!args || !format || !*format)
		return -1;

	pqt_fqtn(fqtn, sizeof(fqtn), args->typhandler->typschema,
		args->typhandler->typname);

	/* put the header */
	PQseterror("%s[pos:%d] - ", fqtn, args->typpos);

	/* append message from type handler */
	va_start(ap, format);
	vseterror(format, ap, TRUE);
	return -1;
}

static void
vseterror(const char *format, va_list ap, int append)
{
	int n;
	int curlen=0;
	char *p;
	va_list ap2;
	pqterr_t *err = geterr();

	if (!err)
		return;

	if (append)
		curlen = (int) strlen(err->msg);
	else
		*err->msg = 0;

  while (1)
	{
		va_copy(ap2, ap);
		n = pqt_vsnprintf(err->msg+curlen, err->size-curlen, format, ap2);
		va_end(ap2);

		/* success */
		if (n > -1)
			break;

		/* need more space */
		n = err->size * 2;
		if (!(p = pqt_realloc(err->msg, n)))
		{
			*err->msg = 0;
			break;
		}

		err->msg = p;
		err->size = n;
	}
}

